import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, box, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.utils import my_round
from vgrid.generator.rhealpixgrid import fix_rhealpix_antimeridian_cells
from vgrid.generator.settings import geodesic_dggs_to_feature
from vgrid.conversion.dggscompact import rhealpix_compact
from vgrid.utils.geometry import check_predicate
import math
from vgrid.stats.rhealpixstats import rhealpix_metrics
from vgrid.utils.geometry import (
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)


def validate_rhealpix_resolution(resolution):
    """
    Validate that rHEALPix resolution is in the valid range [0..15].

    Args:
        resolution: Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [0..15]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 0 or resolution > 15:
        raise ValueError(f"Resolution must be in range [0..15], got {resolution}")

    return resolution


def rhealpix_cell_to_polygon(cell):
    """Convert rHEALPix cell to Shapely polygon.
    
    Args:
        cell: rHEALPix cell object
        
    Returns:
        Polygon: Shapely polygon representation of the cell
    """
    vertices = [
        tuple(my_round(coord, 14) for coord in vertex)
        for vertex in cell.vertices(plane=False)
    ]
    if vertices[0] != vertices[-1]:
        vertices.append(vertices[0])
    vertices = fix_rhealpix_antimeridian_cells(vertices)
    return Polygon(vertices)


def point2rhealpix(
    rhealpix_dggs,
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,
):
    """Convert point geometry to rHEALPix grid cells.
    
    Args:
        rhealpix_dggs: RHEALPixDGGS instance
        resolution: rHEALPix resolution level
        point: Point geometry
        feature_properties: Optional properties to include
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        all_points: All points for topology preservation
        
    Returns:
        list: List of rHEALPix feature dictionaries
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        shortest_distance = shortest_point_distance(all_points)
        # print(shortest_distance)
        if shortest_distance > 0:
            for res in range(16):
                _, avg_edge_length, _ = rhealpix_metrics(res)
                cell_diameter = avg_edge_length * 2
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                resolution = 15
    
    rhealpix_features = []
    seed_cell = rhealpix_dggs.cell_from_point(
        resolution, (point.x, point.y), plane=False
    )

    seed_cell_polygon = rhealpix_cell_to_polygon(seed_cell)
    if seed_cell_polygon:
        seed_cell_id = str(seed_cell)
        num_edges = 4
        if seed_cell.ellipsoidal_shape() == "dart":
            num_edges = 3
        rhealpix_feature = geodesic_dggs_to_feature(
            "rhealpix", seed_cell_id, resolution, seed_cell_polygon, num_edges
        )
        if feature_properties:
            rhealpix_feature["properties"].update(feature_properties)
        rhealpix_features.append(rhealpix_feature)
    return rhealpix_features


def polyline2rhealpix(
    rhealpix_dggs,
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,
):
    """Convert polyline geometry to rHEALPix grid cells.
    
    Args:
        rhealpix_dggs: RHEALPixDGGS instance
        resolution: rHEALPix resolution level
        feature: LineString or MultiLineString geometry
        feature_properties: Optional properties to include
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        all_polylines: All polylines for topology preservation
        
    Returns:
        list: List of rHEALPix feature dictionaries
    """
       # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        shortest_distance = shortest_polyline_distance(all_polylines)
        if shortest_distance > 0:
            for res in range(16):
                _, avg_edge_length, _ = rhealpix_metrics(res)
                cell_diameter = avg_edge_length * 2
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                resolution = 15
    rhealpix_features = []
    polylines = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)

    for polyline in polylines:
        minx, miny, maxx, maxy = polyline.bounds
        bbox_polygon = box(minx, miny, maxx, maxy)
        bbox_center_lon = bbox_polygon.centroid.x
        bbox_center_lat = bbox_polygon.centroid.y
        seed_point = (bbox_center_lon, bbox_center_lat)
        seed_cell = rhealpix_dggs.cell_from_point(resolution, seed_point, plane=False)
        seed_cell_id = str(seed_cell)
        seed_cell_polygon = rhealpix_cell_to_polygon(seed_cell)
        if seed_cell_polygon.contains(bbox_polygon):
            num_edges = 4
            if seed_cell.ellipsoidal_shape() == "dart":
                num_edges = 3
            cell_resolution = resolution
            rhealpix_feature = geodesic_dggs_to_feature(
                "rhealpix", seed_cell_id, cell_resolution, seed_cell_polygon, num_edges
            )
            if feature_properties:
                rhealpix_feature["properties"].update(feature_properties)
            rhealpix_features.append(rhealpix_feature)
            return rhealpix_features
        else:
            covered_cells = set()
            queue = [seed_cell]
            while queue:
                current_cell = queue.pop()
                current_cell_id = str(current_cell)
                if current_cell_id in covered_cells:
                    continue
                covered_cells.add(current_cell_id)
                cell_polygon = rhealpix_cell_to_polygon(current_cell)
                if not cell_polygon.intersects(bbox_polygon):
                    continue
                neighbors = current_cell.neighbors(plane=False)
                for _, neighbor in neighbors.items():
                    neighbor_id = str(neighbor)
                    if neighbor_id not in covered_cells:
                        queue.append(neighbor)
            if compact:
                covered_cells = rhealpix_compact(rhealpix_dggs, covered_cells)
            for cell_id in covered_cells:
                rhealpix_uids = (cell_id[0],) + tuple(map(int, cell_id[1:]))
                rhelpix_cell = rhealpix_dggs.cell(rhealpix_uids)
                cell_resolution = rhelpix_cell.resolution
                cell_polygon = rhealpix_cell_to_polygon(rhelpix_cell)
                if not cell_polygon.intersects(polyline):
                    continue
                num_edges = 4
                if seed_cell.ellipsoidal_shape() == "dart":
                    num_edges = 3
                rhealpix_feature = geodesic_dggs_to_feature(
                    "rhealpix", str(cell_id), cell_resolution, cell_polygon, num_edges
                )
                if feature_properties:
                    rhealpix_feature["properties"].update(feature_properties)
                rhealpix_features.append(rhealpix_feature)

 

    return rhealpix_features


def polygon2rhealpix(
    rhealpix_dggs,
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,
):
    """Convert polygon geometry to rHEALPix grid cells.
    
    Args:
        rhealpix_dggs: RHEALPixDGGS instance
        resolution: rHEALPix resolution level
        feature: Polygon or MultiPolygon geometry
        feature_properties: Optional properties to include
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        all_polygons: All polygons for topology preservation
        
    Returns:
        list: List of rHEALPix feature dictionaries
    """
     # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        shortest_distance = shortest_polygon_distance(all_polygons)
        if shortest_distance > 0:
            for res in range(16):
                _, avg_edge_length, _ = rhealpix_metrics(res)
                cell_diameter = avg_edge_length * 2
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                resolution = 15
    rhealpix_features = []
    polygons = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)

    for polygon in polygons:
        minx, miny, maxx, maxy = polygon.bounds
        bbox_polygon = box(minx, miny, maxx, maxy)
        bbox_center_lon = bbox_polygon.centroid.x
        bbox_center_lat = bbox_polygon.centroid.y
        seed_point = (bbox_center_lon, bbox_center_lat)
        seed_cell = rhealpix_dggs.cell_from_point(resolution, seed_point, plane=False)
        seed_cell_id = str(seed_cell)
        seed_cell_polygon = rhealpix_cell_to_polygon(seed_cell)
        if seed_cell_polygon.contains(bbox_polygon):
            num_edges = 4
            if seed_cell.ellipsoidal_shape() == "dart":
                num_edges = 3
            cell_resolution = resolution
            rhealpix_feature = geodesic_dggs_to_feature(
                "rhealpix", seed_cell_id, cell_resolution, seed_cell_polygon, num_edges
            )
            if feature_properties:
                rhealpix_feature["properties"].update(feature_properties)
            rhealpix_features.append(rhealpix_feature)
            return rhealpix_features
        else:
            covered_cells = set()
            queue = [seed_cell]
            while queue:
                current_cell = queue.pop()
                current_cell_id = str(current_cell)
                if current_cell_id in covered_cells:
                    continue
                covered_cells.add(current_cell_id)
                cell_polygon = rhealpix_cell_to_polygon(current_cell)
                if not cell_polygon.intersects(bbox_polygon):
                    continue
                neighbors = current_cell.neighbors(plane=False)
                for _, neighbor in neighbors.items():
                    neighbor_id = str(neighbor)
                    if neighbor_id not in covered_cells:
                        queue.append(neighbor)
            if compact:
                covered_cells = rhealpix_compact(rhealpix_dggs, covered_cells)
            for cell_id in covered_cells:
                rhealpix_uids = (cell_id[0],) + tuple(map(int, cell_id[1:]))
                rhelpix_cell = rhealpix_dggs.cell(rhealpix_uids)
                cell_resolution = rhelpix_cell.resolution
                cell_polygon = rhealpix_cell_to_polygon(rhelpix_cell)
                if not check_predicate(cell_polygon, polygon, predicate):
                    continue
                num_edges = 4
                if seed_cell.ellipsoidal_shape() == "dart":
                    num_edges = 3
                rhealpix_feature = geodesic_dggs_to_feature(
                    "rhealpix", str(cell_id), cell_resolution, cell_polygon, num_edges
                )
                if feature_properties:
                    rhealpix_feature["properties"].update(feature_properties)
                rhealpix_features.append(rhealpix_feature)

    return rhealpix_features


def geometry2rhealpix(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """Convert list of geometries to rHEALPix grid cells.
    
    Args:
        geometries: Single geometry or list of geometries
        resolution: rHEALPix resolution level
        properties_list: Optional list of properties for each geometry
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        
    Returns:
        dict: GeoJSON FeatureCollection with rHEALPix features
    """
    resolution = validate_rhealpix_resolution(resolution)
    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    rhealpix_dggs = RHEALPixDGGS()
    rhealpix_features = []

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    for idx, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue
        props = (
            properties_list[idx]
            if properties_list and idx < len(properties_list)
            else {}
        )
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            rhealpix_features.extend(
                point2rhealpix(
                    rhealpix_dggs,
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                rhealpix_features.extend(
                    point2rhealpix(
                        rhealpix_dggs,
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            rhealpix_features.extend(
                polyline2rhealpix(
                    rhealpix_dggs,
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            rhealpix_features.extend(
                polygon2rhealpix(
                    rhealpix_dggs,
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons,
                )
            )
    return {"type": "FeatureCollection", "features": rhealpix_features}


def dataframe2rhealpix(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """Convert pandas DataFrame with geometry column to rHEALPix grid cells.
    
    Args:
        df: pandas DataFrame with geometry column
        resolution: rHEALPix resolution level
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        
    Returns:
        dict: GeoJSON FeatureCollection with rHEALPix features
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2rhealpix(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2rhealpix(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """Convert GeoDataFrame to rHEALPix grid cells.
    
    Args:
        gdf: GeoDataFrame with geometry column
        resolution: rHEALPix resolution level
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        include_properties: Whether to include properties
        
    Returns:
        dict: GeoJSON FeatureCollection with rHEALPix features
    """
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2rhealpix(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def vector2rhealpix(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """Convert vector data to rHEALPix grid cells.
    
    Args:
        data: Input data (GeoDataFrame, DataFrame, geometry, GeoJSON dict, or file path)
        resolution: rHEALPix resolution level
        predicate: Spatial predicate for filtering
        compact: Enable compact mode
        topology: Enable topology preserving mode
        output_format: Output format (geojson, gpkg, parquet, csv, shapefile)
        output_path: Output file path
        include_properties: Whether to include properties
        **kwargs: Additional arguments for file reading
        
    Returns:
        str or dict: Output file path or GeoJSON FeatureCollection
    """
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        result = geodataframe2rhealpix(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        result = dataframe2rhealpix(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        result = geometry2rhealpix(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2rhealpix(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2rhealpix(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert rHEALPix result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since rHEALPix uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2rhealpix.gpkg", driver="GPKG")
            return "vector2rhealpix.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2rhealpix.parquet", index=False)
            return "vector2rhealpix.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2rhealpix.shp", driver="ESRI Shapefile")
            return "vector2rhealpix.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2rhealpix_cli():
    """Command-line interface for vector2rhealpix conversion."""
    parser = argparse.ArgumentParser(
        description="Convert vector data to rHEALPix grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(16),
        metavar="[0-15]",
        help="rHEALPix resolution [0..15] (0=coarsest, 15=finest)",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable rHEALPix compact mode for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_rhealpix_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2rhealpix{extensions.get(args.format, '')}"
    try:
        vector2rhealpix(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2rhealpix_cli()



import os
import argparse
import json
from tqdm import tqdm
import rasterio
from vgrid.utils import tilecode, mercantile
import numpy as np
from shapely.geometry import Polygon
from vgrid.stats.quadkeystats import quadkey_metrics
from vgrid.generator.settings import graticule_dggs_to_feature
from math import cos, radians
import csv


def get_nearest_quadkey_resolution(raster_path):
    with rasterio.open(raster_path) as src:
        transform = src.transform
        crs = src.crs
        pixel_width = transform.a
        pixel_height = -transform.e
        cell_size = pixel_width * pixel_height

        if crs.is_geographic:
            # Latitude of the raster center
            center_latitude = (src.bounds.top + src.bounds.bottom) / 2
            # Convert degrees to meters
            meter_per_degree_lat = 111_320  # Roughly 1 degree latitude in meters
            meter_per_degree_lon = meter_per_degree_lat * cos(radians(center_latitude))

            pixel_width_m = pixel_width * meter_per_degree_lon
            pixel_height_m = pixel_height * meter_per_degree_lat
            cell_size = pixel_width_m * pixel_height_m

    nearest_resolution = None
    min_diff = float("inf")

    # Check resolutions from 0 to 29
    for res in range(30):
        _, _, avg_area = quadkey_metrics(res)
        diff = abs(avg_area - cell_size)
        # If the difference is smaller than the current minimum, update the nearest resolution
        if diff < min_diff:
            min_diff = diff
            nearest_resolution = res

    return nearest_resolution


def convert_numpy_types(obj):
    """Recursively convert NumPy types to native Python types"""
    if isinstance(obj, np.generic):
        return obj.item()  # Convert numpy types like np.uint8 to native Python int
    elif isinstance(obj, dict):
        return {k: convert_numpy_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_numpy_types(v) for v in obj]
    else:
        return obj


def raster2quadkey(raster_path, resolution=None, format="geojson"):
    """Convert raster to quadkey format

    Args:
        raster_path (str): Path to input raster file
        resolution (int, optional): Quadkey resolution level [0-29]. If None, will be determined automatically
        format (str, optional): Output format, either 'geojson' or 'csv'. Defaults to 'geojson'

    Returns:
        dict: For geojson format, returns a GeoJSON FeatureCollection
        str: For csv format, returns a CSV string
    """
    # Step 1: Determine the nearest quadkey resolution if none is provided
    if resolution is None:
        resolution = get_nearest_quadkey_resolution(raster_path)
        print(f"Nearest quadkey resolution determined: {resolution}")

    # Open the raster file to get metadata and data
    with rasterio.open(raster_path) as src:
        raster_data = src.read()  # Read all bands
        transform = src.transform
        width, height = src.width, src.height
        band_count = src.count  # Number of bands in the raster

    quadkey_ids = set()

    for row in range(height):
        for col in range(width):
            lon, lat = transform * (col, row)
            quadkey_id = tilecode.latlon2quadkey(lat, lon, resolution)
            quadkey_ids.add(quadkey_id)

    # Sample the raster values at the centroids of the quadkey hexagons
    quadkey_data = []

    for quadkey_id in tqdm(quadkey_ids, desc="Resampling", unit=" cells"):
        # Get the centroid of the quadkey cell
        centroid_lat, centroid_lon = tilecode.quadkey2latlon(quadkey_id)

        # Sample the raster values at the centroid (lat, lon)
        col, row = ~transform * (centroid_lon, centroid_lat)

        if 0 <= col < width and 0 <= row < height:
            # Get the values for all bands at this centroid
            values = raster_data[:, int(row), int(col)]
            quadkey_data.append(
                {
                    "quadkey": quadkey_id,
                    **{
                        f"band_{i + 1}": values[i] for i in range(band_count)
                    },  # Create separate columns for each band
                }
            )

    if format == "csv":
        import io

        output = io.StringIO()
        if quadkey_data:
            writer = csv.DictWriter(output, fieldnames=quadkey_data[0].keys())
            writer.writeheader()
            writer.writerows(quadkey_data)
        return output.getvalue()

    # Create the GeoJSON-like structure
    quadkey_features = []
    for data in tqdm(quadkey_data, desc="Converting to GeoJSON", unit=" cells"):
        quadkey_id = data["quadkey"]
        tile = mercantile.quadkey_to_tile(quadkey_id)
        # Format as tilecode_id
        z = tile.z
        x = tile.x
        y = tile.y
        # Get the bounds of the tile in (west, south, east, north)
        bounds = mercantile.bounds(x, y, z)
        # Create the bounding box coordinates for the polygon
        min_lat, min_lon = bounds.south, bounds.west
        max_lat, max_lon = bounds.north, bounds.east
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],  # Bottom-left corner
                [max_lon, min_lat],  # Bottom-right corner
                [max_lon, max_lat],  # Top-right corner
                [min_lon, max_lat],  # Top-left corner
                [min_lon, min_lat],  # Closing the polygon (same as the first point)
            ]
        )

        cell_resolution = z
        quadkey_feature = graticule_dggs_to_feature(
            "quadkey", quadkey_id, cell_resolution, cell_polygon
        )
        band_properties = {
            f"band_{i + 1}": data[f"band_{i + 1}"] for i in range(band_count)
        }
        quadkey_feature["properties"].update(convert_numpy_types(band_properties))
        quadkey_features.append(quadkey_feature)

    return {
        "type": "FeatureCollection",
        "features": quadkey_features,
    }


def raster2quadkey_cli():
    parser = argparse.ArgumentParser(
        description="Convert Raster in Geographic CRS to Quadkey DGGS"
    )
    parser.add_argument("-raster", type=str, required=True, help="Raster file path")

    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        required=False,
        default=None,
        help="Resolution [0..29]",
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=False,
        default="geojson",
        choices=["geojson", "csv"],
        help="Output format: 'geojson' for GeoJSON format with geometry, 'csv' for tabular data with quadkey and band values",
    )

    args = parser.parse_args()
    raster = args.raster
    resolution = args.resolution
    format = args.format.lower()

    if not os.path.exists(raster):
        print(f"Error: The file {raster} does not exist.")
        return

    if resolution is not None:
        if resolution < 0 or resolution > 29:
            print("Please select a resolution in [0..29] range and try again ")
            return

    result = raster2quadkey(raster, resolution, format)

    output_name = os.path.splitext(os.path.basename(raster))[0]
    output_path = f"{output_name}2quadkey.{format}"

    if format.lower() == "csv":
        with open(output_path, "w", newline="") as f:
            f.write(result)
    else:  # geojson
        with open(output_path, "w") as f:
            json.dump(result, f)

    print(f"Output saved as {output_path}")



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.eaggr import Eaggr
from vgrid.utils.eaggr.enums.model import Model
from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
from vgrid.utils.eaggr.exceptions.eaggr_exception import EaggrException
from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint
from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
from vgrid.utils.eaggr.enums.dggs_return_code import DggsReturnCode


## Unit tests for the error handling in the Eaggr class.
class TestDggsErrors(unittest.TestCase):
    def test_convert_shapes_to_dggs_shapes(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input types
        try:
            dggs.convert_shapes_to_dggs_shapes("Not a list")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                (
                    "Argument must be a list containing only LatLongPoint, "
                    "LatLongLinestring and LatLongPolygon objects"
                ),
            )
        try:
            dggs.convert_shapes_to_dggs_shapes(["Not a LatLongShape"])
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                (
                    "Element 0 in the list is not a LatLongPoint, LatLongLinestring "
                    "or LatLongPolygon object"
                ),
            )
        # Points with invalid values
        try:
            dggs.convert_shapes_to_dggs_shapes([LatLongPoint(None, 0.0, 1.0)])
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                (
                    "Latitude, longitude and accuracy values for a LatLongPoint object "
                    "must all be numeric"
                ),
            )
        try:
            dggs.convert_shapes_to_dggs_shapes([LatLongPoint(360.0, 360.0, 0.0)])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Latitude is greater than maximum allowed "
                    "for a lat/long point."
                ),
            )
        # Empty list of shapes (not actually an error)
        lat_long_shapes = dggs.convert_shapes_to_dggs_shapes([])
        self.assertEqual(lat_long_shapes, [])

    def test_convert_shape_string_to_dggs_shapes(self):
        wkt_string = "LINESTRING(2.345 1.234, 4.567 3.456)"
        string_format = ShapeStringFormat.WKT
        accuracy = 3.884
        dggs = Eaggr(Model.ISEA4T)
        # Invalid input types
        try:
            dggs.convert_shape_string_to_dggs_shapes(None, string_format, accuracy)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Invalid shape string")
        try:
            dggs.convert_shape_string_to_dggs_shapes(wkt_string, 2, accuracy)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                "Shape string format must be one of the ShapeStringFormat values",
            )
        try:
            dggs.convert_shape_string_to_dggs_shapes(wkt_string, string_format, None)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Invalid accuracy value")

        # Empty shape string
        try:
            dggs.convert_shape_string_to_dggs_shapes("", string_format, accuracy)
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e), "EAGGR Exception: Unsupported WKT: . WKT string is not valid."
            )
        # Syntax error in shape string
        try:
            dggs.convert_shape_string_to_dggs_shapes(
                "LINESTRING(2.345 1.234, 4.567 X)", string_format, accuracy
            )
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Unsupported WKT: LINESTRING(2.345 1.234, 4.567 X). "
                    "WKT string is not valid."
                ),
            )
        # Invalid accuracy value
        try:
            dggs.convert_shape_string_to_dggs_shapes(wkt_string, string_format, 0.0)
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(str(e), "EAGGR Exception: Accuracy value must be > 0")

    def test_convert_dggs_cells_to_points(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.convert_dggs_cells_to_points("Not a list")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e), "Argument must be a list containing only DggsCell objects"
            )
        try:
            dggs.convert_dggs_cells_to_points(["Not a DggsCell"])
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Element 0 in the list is not a DggsCell object")
        # Invalid cell ID
        try:
            dggs.convert_dggs_cells_to_points([DggsCell("XXXXXX")])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Empty list of cells (not actually an error)
        lat_long_shapes = dggs.convert_dggs_cells_to_points([])
        self.assertEqual(lat_long_shapes, [])

    def test_convert_dggs_cells_to_shape_string(self):
        dggs_cells = [DggsCell("01230123")]
        string_format = ShapeStringFormat.WKT
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.convert_dggs_cells_to_shape_string("Not a list", string_format)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e), "Argument must be a list containing only DggsCell objects"
            )
        try:
            dggs.convert_dggs_cells_to_shape_string(["Not a DggsCell"], string_format)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Element 0 in the list is not a DggsCell object")
        # Invalid cell ID
        try:
            dggs.convert_dggs_cells_to_shape_string([DggsCell("XXXXXX")], string_format)
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Empty list of cells (not actually an error)
        shape_string = dggs.convert_dggs_cells_to_shape_string([], string_format)
        self.assertEqual(shape_string, "MULTIPOINT EMPTY")
        # Invalid shape string format
        try:
            dggs.convert_dggs_cells_to_shape_string(dggs_cells, 2)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                "Shape string format must be one of the ShapeStringFormat values",
            )

    def test_get_dggs_cell_parents(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.get_dggs_cell_parents("Not a DggsCell")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Argument must be a DggsCell object")
        # Invalid cell ID
        try:
            dggs.get_dggs_cell_parents(DggsCell("XXXXXX"))
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Cell representing a face - no parent cell
        try:
            dggs.get_dggs_cell_parents(DggsCell("01"))
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            ## @todo Improve the error message given by the DLL for this error.
            #        Clearly there is a wrap-around issue with the resolution value.
            #        Issue #17 raised for this
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Unable to create cell at resolution 65535 "
                    "as it is greater than the upper limit (40)."
                ),
            )

    def test_get_dggs_cell_children(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.get_dggs_cell_children("Not a DggsCell")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Argument must be a DggsCell object")
        # Invalid cell ID
        try:
            dggs.get_dggs_cell_children(DggsCell("XXXXXX"))
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Children which exceed the maximum resolution
        try:
            dggs.get_dggs_cell_children(
                DggsCell("012301230123012301230123012301230123012301")
            )
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Unable to create cell at resolution 41 "
                    "as it is greater than the upper limit (40)."
                ),
            )

    def test_get_dggs_cell_siblings(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.get_dggs_cell_siblings("Not a DggsCell")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Argument must be a DggsCell object")
        # Invalid cell ID
        try:
            dggs.get_dggs_cell_siblings(DggsCell("XXXXXX"))
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Cell representing a face - no parent cell
        try:
            dggs.get_dggs_cell_parents(DggsCell("01"))
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            ## @todo Improve the error message given by the DLL for this error.
            #        Clearly there is a wrap-around issue with the resolution value.
            #        Issue #17 raised for this
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Unable to create cell at resolution 65535 "
                    "as it is greater than the upper limit (40)."
                ),
            )
        # Cell exceeds maximum resolution
        try:
            dggs.get_dggs_cell_children(
                DggsCell("012301230123012301230123012301230123012301")
            )
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                (
                    "EAGGR Exception: Unable to create cell at resolution 41 "
                    "as it is greater than the upper limit (40)."
                ),
            )

    def test_get_bounding_dggs_cell(self):
        dggs = Eaggr(Model.ISEA4T)
        # Wrong input type
        try:
            dggs.get_bounding_dggs_cell("Not a list")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e), "Argument must be a list containing only DggsCell objects"
            )
        try:
            dggs.get_bounding_dggs_cell(["Not a DggsCell"])
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Element 0 in the list is not a DggsCell object")
        # Invalid cell ID
        try:
            dggs.get_bounding_dggs_cell([DggsCell("XXXXXX")])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Empty list of cells
        try:
            dggs.get_bounding_dggs_cell([])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e), "EAGGR Exception: No cells provided to determine uniqueness"
            )
        # Cells on different faces - no bounding cell
        try:
            dggs.get_bounding_dggs_cell([DggsCell("0100"), DggsCell("0200")])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_INVALID_PARAM)
            self.assertEqual(str(e), "Supplied cells do not have a common parent.")

    def test_create_dggs_kml_file(self):
        dggs = Eaggr(Model.ISEA4T)
        kml_filename = "KmlFile.kml"
        dggs_cells = [DggsCell("0123")]
        # Wrong input types
        try:
            dggs.create_dggs_kml_file(None, dggs_cells)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Invalid filename for KML file")
        try:
            dggs.create_dggs_kml_file(kml_filename, "Not a list")
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e), "Argument must be a list containing only DggsCell objects"
            )
        try:
            dggs.create_dggs_kml_file(kml_filename, ["Not a DggsCell"])
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Element 0 in the list is not a DggsCell object")
        ## @todo Fix the DLL so that this error is handled.
        """
        # Directory does not exist
        try:
            eaggr.create_dggs_kml_file(os.path.join('not', 'a', 'directory'), dggs_cells)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), '???')
        """
        # Invalid cell ID
        try:
            dggs.create_dggs_kml_file(kml_filename, [DggsCell("XXXXXX")])
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )

    def test_get_dggs_cell_shape_string(self):
        dggs_cells = DggsCell("01230123")
        string_format = ShapeStringFormat.WKT
        dggs = Eaggr(Model.ISEA4T)
        # No cell
        try:
            dggs.convert_dggs_cell_outline_to_shape_string(None, string_format)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(str(e), "Argument must be a DggsCell object")
        # Invalid cell ID
        try:
            dggs.convert_dggs_cell_outline_to_shape_string(
                DggsCell("XXXXXX"), string_format
            )
            self._check_exception_thrown()
        except EaggrException as e:
            self.assertEqual(e.get_return_code(), DggsReturnCode.DGGS_MODEL_ERROR)
            self.assertEqual(
                str(e),
                "EAGGR Exception: Invalid face index, 'XX', must be positive integer",
            )
        # Invalid shape string format
        try:
            dggs.convert_dggs_cell_outline_to_shape_string(dggs_cells, 2)
            self._check_exception_thrown()
        except ValueError as e:
            self.assertEqual(
                str(e),
                "Shape string format must be one of the ShapeStringFormat values",
            )

    def _check_exception_thrown(self):
        self.fail("Should not reach this line - code before should throw an exception")



# Reference: https://github.com/paulojraposo/QTM/blob/master/qtmgenerator.py

from shapely.geometry import Polygon, LinearRing
import json
import os


def constructGeometry(facet):
    vertexTuples = facet[:4]
    # Create a LinearRing with the vertices
    ring = LinearRing(
        [(vT[1], vT[0]) for vT in vertexTuples]
    )  # sequence: lon, lat (x,y)

    # Create a Polygon from the LinearRing
    poly = Polygon(ring)
    return poly


def main():
    outFileDir = os.getcwd()  # Use current directory
    p90_n180, p90_n90, p90_p0, p90_p90, p90_p180 = (
        (90.0, -180.0),
        (90.0, -90.0),
        (90.0, 0.0),
        (90.0, 90.0),
        (90.0, 180.0),
    )
    p0_n180, p0_n90, p0_p0, p0_p90, p0_p180 = (
        (0.0, -180.0),
        (0.0, -90.0),
        (0.0, 0.0),
        (0.0, 90.0),
        (0.0, 180.0),
    )
    n90_n180, n90_n90, n90_p0, n90_p90, n90_p180 = (
        (-90.0, -180.0),
        (-90.0, -90.0),
        (-90.0, 0.0),
        (-90.0, 90.0),
        (-90.0, 180.0),
    )

    levelFacets = {}
    QTMID = {}
    levelFacets[0] = []
    QTMID[0] = []
    geojson_features = []  # Store GeoJSON features separately

    outFileName = "octahedron.geojson"
    outFile = os.path.join(outFileDir, outFileName)

    initial_facets = [
        [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
        [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
        [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
        [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
        [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
        [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
        [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
        [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
    ]

    for i, facet in enumerate(initial_facets):
        QTMID[0].append(str(i + 1))
        geojson_features.append(
            {
                "type": "Feature",
                "geometry": constructGeometry(facet).__geo_interface__,
                "properties": {"qtm": QTMID[0][i]},
            }
        )
        levelFacets[0].append(facet)

    with open(outFile, "w") as f:
        json.dump({"type": "FeatureCollection", "features": geojson_features}, f)

    print(f"Octahedron saved as {outFile}")


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint


## Defines a linestring made up of WGS84 lat/long points.
class LatLongLinestring(object):
    ## Initially the linestring contains no points.
    #  @note Call add_point() to add points to the linestring.
    def __init__(self):
        self._points = []

    ## Adds a point to the linestring.
    #  @note Points must be added to the linestring in the correct order.
    #  @param point Point to add to the linestring.
    #  @throw ValueError Thrown if input argument is not a LatLongPoint object.
    def add_point(self, point):
        if not isinstance(point, LatLongPoint):
            raise ValueError("Input argument to add_point() must be a LatLongPoint")
        self._points.append(point)

    ## @return List of points that make up the linestring.
    def get_points(self):
        return self._points



"""Convert vector data to H3 grid cells."""

import sys
import argparse
from tqdm import tqdm
from pyproj import Geod
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon, box,MultiPoint, MultiLineString, MultiPolygon
import h3
from vgrid.generator.h3grid import fix_h3_antimeridian_cells
from vgrid.generator.settings import geodesic_dggs_to_feature
from vgrid.utils.geometry import (
    geodesic_buffer,
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
    )

geod = Geod(ellps="WGS84")

def validate_h3_resolution(resolution):
    """
    Validate that H3 resolution is in the valid range [0..15].

    Args:
        resolution: Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [0..15]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 0 or resolution > 15:
        raise ValueError(f"Resolution must be in range [0..15], got {resolution}")

    return resolution


# Function to generate grid for Point
def point2h3(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,  # New parameter for topology preservation
):
    """
    Convert a point to H3 grid cells.

    Args:
        resolution (int): H3 resolution [0..15]
        point: Shapely Point geometry
        feature_properties (dict): Properties to add to the H3 features
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode
        topology (bool): Enable H3 topology preserving mode - ensures disjoint points have disjoint H3 cells
        include_properties (bool): If False, do not include original feature properties
        all_points: List of all points for topology preservation (required when topology=True)

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        
        # Calculate the shortest distance between all points
        shortest_distance = shortest_point_distance(all_points)
        
        # Find resolution where H3 cell size is smaller than shortest distance
        # This ensures disjoint points have disjoint H3 cells
        if shortest_distance > 0:
            for res in range(16):
                avg_edge_length = h3.average_hexagon_edge_length(res=res, unit="m")
                # Use a factor to ensure sufficient separation (hexagon diameter is ~2x edge length)
                hexagon_diameter = avg_edge_length*2 
                if hexagon_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single point or no distance, use provided resolution
            pass

    h3_features = []
    # Convert point to the seed cell
    h3_id = h3.latlng_to_cell(point.y, point.x, resolution)

    cell_boundary = h3.cell_to_boundary(h3_id)
    # Wrap and filter the boundary
    filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
    # Reverse lat/lon to lon/lat for GeoJSON compatibility
    reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
    cell_polygon = Polygon(reversed_boundary)
    if cell_polygon:
        num_edges = 6
        if h3.is_pentagon(h3_id):
            num_edges = 5
        h3_feature = geodesic_dggs_to_feature(
            "h3", h3_id, resolution, cell_polygon, num_edges
        )
        if include_properties and feature_properties:
            h3_feature["properties"].update(feature_properties)
        h3_features.append(h3_feature)

    return {
        "type": "FeatureCollection",
        "features": h3_features,
    }

def polyline2h3(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,  # New parameter for topology preservation
):
    """
    Convert a polyline to H3 grid cells.

    Args:
        resolution (int): H3 resolution [0..15]
        feature: Shapely LineString or MultiLineString geometry
        feature_properties (dict): Properties to add to the H3 features
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode
        topology (bool): Enable H3 topology preserving mode - ensures disjoint polylines have disjoint H3 cells
        include_properties (bool): If False, do not include original feature properties
        all_polylines: List of all polylines for topology preservation (required when topology=True)

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        
        # Calculate the shortest distance between all polylines
        shortest_distance = shortest_polyline_distance(all_polylines)
        
        # Find resolution where H3 cell size is smaller than shortest distance
        # This ensures disjoint polylines have disjoint H3 cells
        if shortest_distance > 0:
            for res in range(16):
                avg_edge_length = h3.average_hexagon_edge_length(res=res, unit="m")
                # Use a factor to ensure sufficient separation (hexagon diameter is ~2x edge length)
                hexagon_diameter = avg_edge_length * 4 # in case there are 2 cells representing the same line segment
                if hexagon_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single polyline or no distance, use provided resolution
            pass

    h3_features = []
    if feature.geom_type == "LineString":
        polylines = [feature]
    elif feature.geom_type == "MultiLineString":
        polylines = list(feature.geoms)
    else:
        return []
    for polyline in polylines:
        bbox = box(*polyline.bounds)
        distance = h3.average_hexagon_edge_length(resolution, unit="m") * 2
        bbox_buffer = geodesic_buffer(bbox, distance)
        bbox_buffer_cells = h3.geo_to_cells(bbox_buffer, resolution)
        if compact:
            bbox_buffer_cells = h3.compact_cells(bbox_buffer_cells)

        for bbox_buffer_cell in bbox_buffer_cells:
            cell_boundary = h3.cell_to_boundary(bbox_buffer_cell)
            filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
            reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
            cell_polygon = Polygon(reversed_boundary)

            # Use the check_predicate function to determine if we should keep this cell
            if not check_predicate(cell_polygon, polyline, "intersects"):
                continue  # Skip non-matching cells

            cell_resolution = h3.get_resolution(bbox_buffer_cell)
            num_edges = 6
            if h3.is_pentagon(bbox_buffer_cell):
                num_edges = 5
            h3_feature = geodesic_dggs_to_feature(
                "h3", bbox_buffer_cell, cell_resolution, cell_polygon, num_edges
            )
            if include_properties and feature_properties:
                h3_feature["properties"].update(feature_properties)
            h3_features.append(h3_feature)

    return {
        "type": "FeatureCollection",
        "features": h3_features,
    }


def polygon2h3(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,  # New parameter for topology preservation
):
    """
    Convert a polygon to H3 grid cells.

    Args:
        resolution (int): H3 resolution [0..15]
        feature: Shapely Polygon or MultiPolygon geometry
        feature_properties (dict): Properties to add to the H3 features
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode
        topology (bool): Enable H3 topology preserving mode - ensures disjoint polygons have disjoint H3 cells
        include_properties (bool): If False, do not include original feature properties
        all_polygons: List of all polygons for topology preservation (required when topology=True)

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        
       
        # Calculate the shortest distance between all polygons
        shortest_distance = shortest_polygon_distance(all_polygons)
        
        # Find resolution where H3 cell size is smaller than shortest distance
        # This ensures disjoint polygons have disjoint H3 cells
        if shortest_distance > 0:
            for res in range(16):
                avg_edge_length = h3.average_hexagon_edge_length(res=res, unit="m")
                # Use a factor to ensure sufficient separation (hexagon diameter is ~2x edge length)
                hexagon_diameter = avg_edge_length * 4
                if hexagon_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single polygon or no distance, use provided resolution
            pass

    h3_features = []
    if feature.geom_type == "Polygon":
        polygons = [feature]
    elif feature.geom_type == "MultiPolygon":
        polygons = list(feature.geoms)
    else:
        return []
    
    for polygon in polygons:
        bbox = box(*polygon.bounds)
        distance = h3.average_hexagon_edge_length(resolution, unit="m") * 2
        bbox_buffer = geodesic_buffer(bbox, distance)
        bbox_buffer_cells = h3.geo_to_cells(bbox_buffer, resolution)
        if compact:
            bbox_buffer_cells = h3.compact_cells(bbox_buffer_cells)

        for bbox_buffer_cell in bbox_buffer_cells:
            cell_boundary = h3.cell_to_boundary(bbox_buffer_cell)
            filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
            reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
            cell_polygon = Polygon(reversed_boundary)

            # Use the check_predicate function to determine if we should keep this cell
            if not check_predicate(cell_polygon, polygon, predicate):
                continue  # Skip non-matching cells

            cell_resolution = h3.get_resolution(bbox_buffer_cell)
            num_edges = 6
            if h3.is_pentagon(bbox_buffer_cell):
                num_edges = 5
            h3_feature = geodesic_dggs_to_feature(
                "h3", bbox_buffer_cell, cell_resolution, cell_polygon, num_edges
            )
            if include_properties and feature_properties:
                h3_feature["properties"].update(feature_properties)
            h3_features.append(h3_feature)

    return {
        "type": "FeatureCollection",
        "features": h3_features,
    }


def geometry2h3(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert Shapely geometry objects directly to H3 grid cells without converting to GeoJSON first.

    Args:
        geometries: Single Shapely geometry or list of Shapely geometries
        resolution (int): H3 resolution [0..15]
        properties_list: List of property dictionaries (optional)
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode - for polygon only
        topology (bool): Enable H3 topology preserving mode
        include_properties (bool): If False, do not include original feature properties

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    resolution = validate_h3_resolution(resolution)

    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    h3_features = []

    for i, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue

        # Get properties for this geometry
        props = properties_list[i] if i < len(properties_list) else {}

        # Process based on geometry type
        if geom.geom_type == "Point":
            point_features = point2h3(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_points,  # Pass all points for topology preservation
            )
            h3_features.extend(point_features["features"])

        elif geom.geom_type == "MultiPoint":
            for point in geom.geoms:
                point_features = point2h3(
                    resolution,
                    point,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
                h3_features.extend(point_features["features"])

        elif geom.geom_type in ["LineString", "MultiLineString"]:
            polyline_features = polyline2h3(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_polylines,  # Pass all polylines for topology preservation
            )
            h3_features.extend(polyline_features["features"])

        elif geom.geom_type in ["Polygon", "MultiPolygon"]:
            poly_features = polygon2h3(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_polygons=all_polygons,  # Pass all polygons for topology preservation
            )
            h3_features.extend(poly_features["features"])

        else:
            raise ValueError(f"Unsupported geometry type: {geom.geom_type}")

    return {
        "type": "FeatureCollection",
        "features": h3_features,
    }


def dataframe2h3(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert pandas DataFrame with geometry column to H3 grid cells by converting to Shapely geometries first.

    Args:
        df (pd.DataFrame): Input DataFrame with geometry column
        resolution (int): H3 resolution [0..15]
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode - for polygon only
        topology (bool): Enable H3 topology preserving mode
        include_properties (bool): If False, do not include original feature properties

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    # Find geometry column
    geometry_col = None
    for col in df.columns:
        if hasattr(df[col].iloc[0], "geom_type") or hasattr(
            df[col].iloc[0], "__geo_interface__"
        ):
            geometry_col = col
            break

    if geometry_col is None:
        raise ValueError(
            "DataFrame must contain a geometry column with Shapely geometry objects"
        )

    # Extract geometries and properties from DataFrame
    geometries = []
    properties_list = []

    for _, row in df.iterrows():
        # Get the geometry
        geom = row[geometry_col]
        if geom is not None:
            geometries.append(geom)

            # Get properties (exclude geometry column)
            properties = row.to_dict()
            if geometry_col in properties:
                del properties[geometry_col]
            properties_list.append(properties)

    # Use geometry2h3 to process the geometries
    return geometry2h3(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2h3(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert GeoDataFrame to H3 grid cells by converting to Shapely geometries first.

    Args:
        gdf (GeoDataFrame): Input GeoDataFrame
        resolution (int): H3 resolution [0..15]
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode - for polygon only
        topology (bool): Enable H3 topology preserving mode
        include_properties (bool): If False, do not include original feature properties

    Returns:
        dict: GeoJSON FeatureCollection containing H3 grid cells
    """
    # Extract geometries and properties from GeoDataFrame
    geometries = []
    properties_list = []

    for _, row in gdf.iterrows():
        # Get the geometry
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)

            # Get properties (exclude geometry column)
            properties = row.to_dict()
            if "geometry" in properties:
                del properties["geometry"]
            properties_list.append(properties)

    # Use geometry2h3 to process the geometries
    return geometry2h3(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def vector2h3(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to H3 grid cells from various input formats.

    This function can handle:
    - File paths (GeoJSON, Shapefile, GeoPackage, KML, GML, etc.) - converted to GeoDataFrame then Shapely geometries
    - URLs (remote files) - converted to GeoDataFrame then Shapely geometries
    - pandas DataFrames with geometry column - converted to Shapely geometries
    - GeoJSON dictionaries - converted to GeoDataFrame then Shapely geometries
    - Shapely geometry objects - processed directly
    - GeoDataFrames - converted to Shapely geometries

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): H3 resolution [0..15]
        predicate (str or int): Spatial predicate to apply (see check_predicate function)
        compact (bool): Enable H3 compact mode for polygons (default: False)
        topology (bool): Enable H3 topology preserving mode (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions

    Returns:
        dict or str: Output in the specified format
    """
    # Process input data directly
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        # GeoDataFrame - convert to Shapely geometries first
        result = geodataframe2h3(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        # Regular DataFrame with geometry column - convert to Shapely geometries first
        result = dataframe2h3(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        # Shapely geometry objects - process directly
        result = geometry2h3(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        # GeoJSON data - convert to GeoDataFrame first, then process
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2h3(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    
    elif isinstance(data, str):
        # File path or URL - use geopandas.read_file directly
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2h3(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")

    # Convert result to specified output format
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert H3 result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since H3 uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2h3.gpkg", driver="GPKG")
            return "vector2h3.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2h3.parquet", index=False)
            return "vector2h3.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2h3.shp", driver="ESRI Shapefile")
            return "vector2h3.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2h3_cli():
    """Command-line interface for vector2h3 conversion."""
    parser = argparse.ArgumentParser(description="Convert vector data to H3 grid cells")
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(16),
        metavar="[0-15]",
        help="H3 resolution [0..15] (0=coarsest, 15=finest)",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable H3 compact mode for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", 
        help="Enable topology preserving mode ensuring disjoint features have disjoint H3 cells by automatically calculating appropriate resolution."
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")

    args = parser.parse_args()

    # Validate resolution if provided
    if args.resolution is not None:
        try:
            args.resolution = validate_h3_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    # Handle input (no stdin support)
    data = args.input

    # Handle output path
    output_path = args.output

    # If output path is not specified but format requires a file, generate default name
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        # Generate default filename based on format
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2h3{extensions.get(args.format, '')}"

    try:
        vector2h3(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )

        if output_path:
            print(f"Output saved to {output_path}")

    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2h3_cli()



# -*- coding: utf-8 -*-

r"""
This Python 3.3 code implements the zenithial orthotriangular (ZOT) map projection as described in [Dutt1991]_.
.. [Dutt1991] Geoffrey Dutton, Zenithial orthotriangular projection,
Auto-Carto 10 (1991), ACSM-ASPRS Annual Convention, Baltimore, Technical Papers, v. 6, 77--95.
CHANGELOG:
- Alexander Raichev (AR), 2013-01-30: Initial version.
- AR, 2013-07-23: Ported to Python 3.3.
NOTE:
All lengths are measured in meters and all angles are measured in radians
unless indicated otherwise.
By 'ellipsoid' below, i mean an oblate ellipsoid of revolution.
"""
# *****************************************************************************
#       Copyright (C) 2013 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import standard modules.
from math import pi

FLOPS = [1, 1, -1, -1, -1, 1, 1, -1]
P2 = pi / 2


def zot_sphere(lam, phi, diam=1, flops=FLOPS):
    r"""
    Compute the signature function of the ZOT projection of the unit sphere.

    INPUT:

    - `lam, phi` - Geodetic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.

    EXAMPLES::
    """
    s = diam / pi
    org = int((P2 - phi) // P2)
    octant = int((org + 1) * (lam // P2))
    x1 = 2 - ((octant + org - 1) % 2)
    x2 = 3 - x1
    hs = 1 - (2 * org)
    r = (s * FLOPS[octant], -s * hs * FLOPS[9 - octant])
    c = (-org * r[x1], -org * r[x2])
    clp = P2 - (hs * phi)
    olp = clp * (lam // P2) / P2
    px = c[x1] + r[x1] * abs(clp - olp)
    py = c[x2] + r[x2] * olp
    return px, py


def zot_sphere_inverse(x, y):
    r"""
    Compute the inverse of the zot_sphere().

    INPUT:

    - `x, y` - Planar coordinates in meters in the image of the
      ZOT projection of the unit sphere.

    EXAMPLES::
    """
    pass


def zot_ellipsoid(lam, phi, e=0.5):
    pass


def zot_ellipsoid_inverse(x, y, e=0.5):
    pass



import argparse
import json
import os
from shapely.geometry import Polygon
from tqdm import tqdm
from shapely.geometry import shape, box, Point, LineString
from pyproj import Geod
import platform
import geopandas as gpd
import pandas as pd

if platform.system() == "Linux":
    from vgrid.utils.dggrid4py import DGGRIDv7, dggs_types
    from vgrid.utils.dggrid4py.dggrid_runner import output_address_types

geod = Geod(ellps="WGS84")


# Function to generate grid for Point
def point_to_grid(dggrid_instance, dggs_type, res, address_type, geometry):
    """
    Generate DGGRID cell(s) for a Point or MultiPoint geometry.

    Args:
        dggrid_instance: DGGRIDv7 instance for grid operations.
        dggs_type (str): Type of DGGS (e.g., ISEA4H, FULLER, etc.).
        res (int): Resolution for the DGGRID.
        address_type (str): Address type for the output grid cells.
        geometry (shapely.geometry.Point or MultiPoint): Input geometry.

    Returns:
        str: GeoJSON FeatureCollection as a string containing the grid cell(s).
    """
    # Initialize an empty list to store filtered grid cells
    merged_grids = []

    # Check the geometry type
    if geometry.geom_type == "Point":
        # Handle single Point
        points = [geometry]
    elif geometry.geom_type == "MultiPoint":
        # Handle MultiPoint: process each point separately
        points = list(geometry)

    # Process each point
    for point in points:
        # Create a GeoDataFrame for the point in EPSG:4326 CRS
        geodf_points_wgs84 = gpd.GeoDataFrame([{"geometry": point}], crs="EPSG:4326")

        # Get DGGRID cell ID for the point
        dggrid_cell = dggrid_instance.cells_for_geo_points(
            geodf_points_wgs84=geodf_points_wgs84,
            cell_ids_only=True,
            dggs_type=dggs_type,
            resolution=res,
        )

        dggrid_seqnum = dggrid_cell.loc[0, "seqnums"]  # Use 'seqnums' column by default
        # Get the polygon representation of the cell using the cell ID
        dggrid_cell = dggrid_instance.grid_cell_polygons_from_cellids(
            [dggrid_seqnum],
            dggs_type=dggs_type,
            resolution=res,
            split_dateline=True,
            clip_cell_res=1,
            input_address_type="SEQNUM",
            # output_address_type=address_type
        )

        # Ensure the geometry column is set
        gdf = dggrid_cell.set_geometry("geometry")

        # Ensure the CRS is set to EPSG:4326
        if gdf.crs is None:
            gdf.set_crs(epsg=4326, inplace=True)
        elif not gdf.crs.equals("EPSG:4326"):
            gdf = gdf.to_crs(epsg=4326)

        try:
            if address_type != "SEQNUM":
                address_type_transform = dggrid_instance.address_transform(
                    [dggrid_seqnum],
                    dggs_type=dggs_type,
                    resolution=res,
                    mixed_aperture_level=None,
                    input_address_type="SEQNUM",
                    output_address_type=address_type,
                )
                gdf["name"] = gdf["name"].astype(
                    str
                )  # Convert the column to string type
                gdf.loc[0, "name"] = address_type_transform.loc[0, address_type]
                gdf = gdf.rename(columns={"name": address_type.lower()})
            else:
                gdf = gdf.rename(columns={"name": "seqnum"})
        except Exception:
            pass

        # Append the filtered GeoDataFrame to the list
        merged_grids.append(gdf)

    # Merge all filtered grids into one GeoDataFrame
    if merged_grids:
        final_grid = gpd.GeoDataFrame(
            pd.concat(merged_grids, ignore_index=True), crs=merged_grids[0].crs
        )
    else:
        final_grid = gpd.GeoDataFrame(columns=["geometry"], crs="EPSG:4326")

    # Convert the GeoDataFrame to a GeoJSON feature collection
    feature_collection = final_grid.to_json()
    return feature_collection


# Function to generate grid for Polyline
def polyline_to_grid(dggrid_instance, dggs_type, res, address_type, geometry):
    """
    Generate DGGRID cells intersecting with a LineString or MultiLineString geometry.

    Args:
        dggrid_instance: DGGRIDv7 instance for grid operations.
        dggs_type (str): Type of DGGS (e.g., ISEA4H, FULLER, etc.).
        res (int): Resolution for the DGGRID.
        address_type (str): Address type for the output grid cells.
        geometry (shapely.geometry.LineString or MultiLineString): Input geometry.

    Returns:
        geopandas.GeoDataFrame: GeoDataFrame containing DGGRID cells intersecting with the input geometry.
    """
    # Initialize an empty list to store filtered grid cells
    merged_grids = []

    # Check the geometry type
    if geometry.geom_type == "LineString":
        # Handle single LineString
        polylines = [geometry]
    elif geometry.geom_type == "MultiLineString":
        # Handle MultiLineString: process each line separately
        polylines = list(geometry.geoms)

    # Process each polyline
    for polyline in polylines:
        # Get bounding box for the current polyline
        bounding_box = box(*polyline.bounds)

        # Generate grid cells for the bounding box
        dggrid_gdf = dggrid_instance.grid_cell_polygons_for_extent(
            dggs_type,
            res,
            clip_geom=bounding_box,
            split_dateline=True,
            output_address_type=address_type,
        )

        # Keep only grid cells that intersect the polyline
        dggrid_gdf = dggrid_gdf[dggrid_gdf.intersects(polyline)]

        try:
            if address_type != "SEQNUM":

                def address_transform(
                    dggrid_seqnum, dggs_type, resolution, address_type
                ):
                    address_type_transform = dggrid_instance.address_transform(
                        [dggrid_seqnum],
                        dggs_type=dggs_type,
                        resolution=resolution,
                        mixed_aperture_level=None,
                        input_address_type="SEQNUM",
                        output_address_type=address_type,
                    )
                    return address_type_transform.loc[0, address_type]

                dggrid_gdf["name"] = dggrid_gdf["name"].astype(str)
                dggrid_gdf["name"] = dggrid_gdf["name"].apply(
                    lambda val: address_transform(val, dggs_type, res, address_type)
                )
                dggrid_gdf = dggrid_gdf.rename(columns={"name": address_type.lower()})
            else:
                dggrid_gdf = dggrid_gdf.rename(columns={"name": "seqnum"})

        except Exception:
            pass
        # Append the filtered GeoDataFrame to the list
        merged_grids.append(dggrid_gdf)

    # Merge all filtered grids into one GeoDataFrame
    if merged_grids:
        final_grid = gpd.GeoDataFrame(
            pd.concat(merged_grids, ignore_index=True), crs=merged_grids[0].crs
        )
    else:
        final_grid = gpd.GeoDataFrame(columns=["geometry"], crs="EPSG:4326")

    return final_grid


def polygon_to_grid(dggrid_instance, dggs_type, res, address_type, geometry):
    """
    Generate DGGRID cells intersecting with a given polygon or multipolygon geometry.

    Args:
        dggrid_instance: DGGRIDv7 instance for grid operations.
        dggs_type (str): Type of DGGS (e.g., ISEA4H, FULLER, etc.).
        res (int): Resolution for the DGGRID.
        address_type (str): Address type for the output grid cells.
        geometry (shapely.geometry.Polygon or MultiPolygon): Input geometry.

    Returns:
        geopandas.GeoDataFrame: GeoDataFrame containing DGGRID cells intersecting with the input geometry.
    """
    # Initialize an empty list to store filtered grid cells
    merged_grids = []

    # Check the geometry type
    if geometry.geom_type == "Polygon":
        # Handle single Polygon
        polygons = [geometry]
    elif geometry.geom_type == "MultiPolygon":
        # Handle MultiPolygon: process each polygon separately
        polygons = list(geometry.geoms)  # Use .geoms to get components of MultiPolygon

    # Process each polygon
    for polygon in polygons:
        # Get bounding box for the current polygon
        bounding_box = box(*polygon.bounds)

        # Generate grid cells for the bounding box
        dggrid_gdf = dggrid_instance.grid_cell_polygons_for_extent(
            dggs_type,
            res,
            clip_geom=bounding_box,
            split_dateline=True,
            output_address_type=address_type,
        )

        # Keep only grid cells that intersect the polygon
        dggrid_gdf = dggrid_gdf[dggrid_gdf.intersects(polygon)]
        try:
            if address_type != "SEQNUM":

                def address_transform(
                    dggrid_seqnum, dggs_type, resolution, address_type
                ):
                    address_type_transform = dggrid_instance.address_transform(
                        [dggrid_seqnum],
                        dggs_type=dggs_type,
                        resolution=resolution,
                        mixed_aperture_level=None,
                        input_address_type="SEQNUM",
                        output_address_type=address_type,
                    )
                    return address_type_transform.loc[0, address_type]

                dggrid_gdf["name"] = dggrid_gdf["name"].astype(str)
                dggrid_gdf["name"] = dggrid_gdf["name"].apply(
                    lambda val: address_transform(val, dggs_type, res, address_type)
                )
                dggrid_gdf = dggrid_gdf.rename(columns={"name": address_type.lower()})
            else:
                dggrid_gdf = dggrid_gdf.rename(columns={"name": "seqnum"})

        except Exception:
            pass

        # Append the filtered GeoDataFrame to the list
        merged_grids.append(dggrid_gdf)

    # Merge all filtered grids into one GeoDataFrame
    if merged_grids:
        final_grid = gpd.GeoDataFrame(
            pd.concat(merged_grids, ignore_index=True), crs=merged_grids[0].crs
        )
    else:
        final_grid = gpd.GeoDataFrame(columns=["geometry"], crs="EPSG:4326")

    return final_grid


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert DGGRID result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since DGGRID uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2dggrid.gpkg", driver="GPKG")
            return "vector2dggrid.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2dggrid.parquet", index=False)
            return "vector2dggrid.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2dggrid.shp", driver="ESRI Shapefile")
            return "vector2dggrid.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def main():
    """
    Command-line interface for converting GeoJSON to DGGRID grid cells.

    Usage:
        python vector2dggrid.py -t <dggs_type> -r <resolution> -a <address_type> -geojson <geojson_file>

    Arguments:
        -t, --dggs_type: DGGS type (e.g., ISEA4H, FULLER, etc.)
        -r, --resolution: Resolution for the DGGRID
        -a, --address_type: Output address type (default: SEQNUM)
        -geojson, --geojson: Path to GeoJSON file with Point, Polyline, or Polygon

    The function reads the input GeoJSON, processes each feature, and writes the output as a new GeoJSON file with grid cells.
    """
    if platform.system() == "Linux":
        parser = argparse.ArgumentParser(description="Convert GeoJSON to DGGRID")
        parser.add_argument(
            "-t",
            "--dggs_type",
            choices=dggs_types,
            help="Select a DGGS type from the available options.",
        )
        parser.add_argument(
            "-r", "--resolution", type=int, required=True, help="resolution"
        )
        parser.add_argument(
            "-a",
            "--address_type",
            choices=output_address_types,
            default="SEQNUM",
            nargs="?",  # This makes the argument optional
            help="Select an output address type from the available options.",
        )

        parser.add_argument(
            "-geojson",
            "--geojson",
            type=str,
            required=True,
            help="GeoJSON string with Point, Polyline or Polygon",
        )
        dggrid_instance = DGGRIDv7(
            executable="/usr/local/bin/dggrid",
            working_dir=".",
            capture_logs=False,
            silent=True,
            tmp_geo_out_legacy=False,
            debug=False,
        )
        args = parser.parse_args()
        dggs_type = args.dggs_type
        resolution = args.resolution
        address_type = args.address_type
        geojson = args.geojson

        if not os.path.exists(geojson):
            print(f"Error: The file {geojson} does not exist.")
            return

        with open(geojson, "r", encoding="utf-8") as f:
            geojson_data = json.load(f)

            # Initialize a list to store all grid cells
        all_cells = []
        geojson_features = []
        for feature in tqdm(geojson_data["features"], desc="Processing features"):
            if feature["geometry"]["type"] in ["Point", "MultiPoint"]:
                coordinates = feature["geometry"]["coordinates"]

                if feature["geometry"]["type"] in ["Point", "MultiPoint"]:
                    coordinates = feature["geometry"]["coordinates"]

                    if feature["geometry"]["type"] == "Point":
                        # Create a Point geometry
                        point = Point(coordinates)
                        point_features_json = point_to_grid(
                            dggrid_instance, dggs_type, resolution, address_type, point
                        )

                        # Parse the GeoJSON string if it's returned as a string
                        point_features = json.loads(point_features_json)

                        # Assuming point_features is a GeoJSON with 'features'
                        geojson_features.extend(point_features["features"])

                    elif feature["geometry"]["type"] == "MultiPoint":
                        # Process each coordinate set in MultiPoint
                        for point_coords in coordinates:
                            point = Point(
                                point_coords
                            )  # Create Point for each coordinate set
                            point_features_json = point_to_grid(
                                dggrid_instance,
                                dggs_type,
                                resolution,
                                address_type,
                                point,
                            )

                            # Parse the GeoJSON string if it's returned as a string
                            point_features = json.loads(point_features_json)

                            # Append the point features to the list
                            geojson_features.extend(point_features["features"])

            elif feature["geometry"]["type"] in ["LineString", "MultiLineString"]:
                coordinates = feature["geometry"]["coordinates"]
                if feature["geometry"]["type"] == "LineString":
                    # Directly process LineString geometry
                    polyline = LineString(coordinates)
                    cells_gdf = polyline_to_grid(
                        dggrid_instance, dggs_type, resolution, address_type, polyline
                    )
                    all_cells.append(cells_gdf)
                elif feature["geometry"]["type"] == "MultiLineString":
                    # Process each LineString in MultiLineString
                    multilines = shape(feature["geometry"])
                    cells_gdf = polyline_to_grid(
                        dggrid_instance, dggs_type, resolution, address_type, multilines
                    )
                    all_cells.append(cells_gdf)

            if feature["geometry"]["type"] in ["Polygon", "MultiPolygon"]:
                coordinates = feature["geometry"]["coordinates"]
                if feature["geometry"]["type"] == "Polygon":
                    # Extract exterior and interior rings
                    exterior_ring = coordinates[
                        0
                    ]  # The first coordinate set is the exterior ring
                    interior_rings = coordinates[
                        1:
                    ]  # Remaining coordinate sets are interior rings (holes)
                    polygon = Polygon(exterior_ring, interior_rings)

                    # Process the polygon with interior rings
                    cells_gdf = polygon_to_grid(
                        dggrid_instance, dggs_type, resolution, address_type, polygon
                    )
                    all_cells.append(cells_gdf)

                elif feature["geometry"]["type"] == "MultiPolygon":
                    # Process each polygon in MultiPolygon
                    multipolygon = shape(
                        feature["geometry"]
                    )  # Convert MultiPolygon from GeoJSON to Shapely object

                    for (
                        polygon
                    ) in multipolygon.geoms:  # Iterate over individual polygons
                        # Extract exterior and interior rings
                        exterior_ring = polygon.exterior.coords
                        interior_rings = [ring.coords for ring in polygon.interiors]
                        polygon_with_holes = Polygon(exterior_ring, interior_rings)

                        # Process the polygon with interior rings
                        cells_gdf = polygon_to_grid(
                            dggrid_instance,
                            dggs_type,
                            resolution,
                            address_type,
                            polygon_with_holes,
                        )
                        all_cells.append(cells_gdf)

        # Merge all collected GeoDataFrames
        geojson_path = f"geojson2dggrid_{dggs_type}_{resolution}_{address_type}.geojson"

        if geojson_features:
            with open(geojson_path, "w") as f:
                json.dump(
                    {"type": "FeatureCollection", "features": geojson_features}, f
                )
        elif all_cells:
            final_gdf = gpd.GeoDataFrame(
                pd.concat(all_cells, ignore_index=True), crs=all_cells[0].crs
            )
            final_gdf.to_file(geojson_path, driver="GeoJSON")

        print(f"DGGRID GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



# Reference:
# https://github.com/aaliddell/s2cell,
# https://medium.com/@claude.ducharme/selecting-a-geo-representation-81afeaf3bf01
# https://github.com/sidewalklabs/s2
# https://github.com/google/s2geometry/tree/master/src/python
# https://github.com/google/s2geometry
# https://gis.stackexchange.com/questions/293716/creating-shapefile-of-s2-cells-for-given-level
# https://s2.readthedocs.io/en/latest/quickstart.html
from vgrid.utils import s2
import json
import argparse
from tqdm import tqdm
from vgrid.utils.antimeridian import fix_polygon
from shapely.geometry import Polygon
from vgrid.generator.settings import geodesic_dggs_to_feature
from shapely.geometry import shape
from shapely.ops import unary_union


def s2_cell_to_polygon(cell_id):
    cell = s2.Cell(cell_id)
    vertices = []
    for i in range(4):
        vertex = s2.LatLng.from_point(cell.get_vertex(i))
        vertices.append((vertex.lng().degrees, vertex.lat().degrees))

    vertices.append(vertices[0])  # Close the polygon

    # Create a Shapely Polygon
    polygon = Polygon(vertices)
    #  Fix Antimerididan:
    fixed_polygon = fix_polygon(polygon)
    return fixed_polygon


def generate_grid(resolution, bbox):
    min_lng, min_lat, max_lng, max_lat = bbox
    # Define the cell level (S2 uses a level system for zoom, where level 30 is the highest resolution)
    level = resolution
    # Create a list to store the S2 cell IDs
    cell_ids = []
    # Define the cell covering
    coverer = s2.RegionCoverer()
    coverer.min_level = level
    coverer.max_level = level
    # coverer.max_cells = 1000_000  # Adjust as needed
    # coverer.max_cells = 0  # Adjust as needed

    # Define the region to cover (in this example, we'll use the entire world)
    region = s2.LatLngRect(
        s2.LatLng.from_degrees(min_lat, min_lng),
        s2.LatLng.from_degrees(max_lat, max_lng),
    )

    # Get the covering cells
    covering = coverer.get_covering(region)

    # Convert the covering cells to S2 cell IDs
    for cell_id in covering:
        cell_ids.append(cell_id)

    s2_features = []
    num_edges = 4

    for cell_id in tqdm(cell_ids, desc="Generating S2 DGGS", unit=" cells"):
        # Generate a Shapely Polygon
        cell_polygon = s2_cell_to_polygon(cell_id)
        s2_token = cell_id.to_token()
        s2_feature = geodesic_dggs_to_feature(
            "s2", s2_token, resolution, cell_polygon, num_edges
        )
        s2_features.append(s2_feature)

    return {"type": "FeatureCollection", "features": s2_features}


def generate_grid_resample(resolution, geojson_features):
    geometries = [
        shape(feature["geometry"]) for feature in geojson_features["features"]
    ]
    unified_geom = unary_union(geometries)

    # Step 2: Get bounding box from unified geometry
    min_lng, min_lat, max_lng, max_lat = unified_geom.bounds

    # Step 3: Configure the S2 coverer
    level = resolution
    coverer = s2.RegionCoverer()
    coverer.min_level = level
    coverer.max_level = level

    # Step 4: Create a LatLngRect from the bounding box
    region = s2.LatLngRect(
        s2.LatLng.from_degrees(min_lat, min_lng),
        s2.LatLng.from_degrees(max_lat, max_lng),
    )

    # Step 5: Get the covering cells
    covering = coverer.get_covering(region)

    s2_features = []
    for cell_id in tqdm(covering, desc="Generating S2 DGGS", unit=" cells"):
        # Convert S2 cell to polygon (must define `s2_cell_to_polygon`)
        cell_polygon = s2_cell_to_polygon(cell_id)

        # Check intersection with actual geometry
        if cell_polygon.intersects(unified_geom):
            s2_token = cell_id.to_token()
            num_edges = 4
            s2_feature = geodesic_dggs_to_feature(
                "s2", s2_token, resolution, cell_polygon, num_edges
            )
            s2_features.append(s2_feature)

    return {"type": "FeatureCollection", "features": s2_features}


def main():
    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Generate S2 DGGS.")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="Resolution [0..30]"
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )
    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox if args.bbox else [-180, -90, 180, 90]

    if resolution < 0 or resolution > 30:
        print("Please select a resolution in [0..30] range and try again ")
        return

    geojson_features = generate_grid(resolution, bbox)
    # Define the GeoJSON file path
    geojson_path = f"s2_grid_{resolution}.geojson"
    with open(geojson_path, "w") as f:
        json.dump(geojson_features, f, indent=2)


if __name__ == "__main__":
    main()



# Reference: https://github.com/paulojraposo/QTM/blob/master/qtmgenerator.py

# -*- coding: utf-8 -*-

#   .-.
#   /v\    L   I   N   U   X
#  // \\
# /(   )\
#  ^^-^^

# This script makes a Quarternary Triangular Mesh (QTM) to tessellate the world based
# on an octohedron, based on Geoffrey Dutton's conception:
#
# Dutton, Geoffrey H. "Planetary Modelling via Hierarchical Tessellation." In Procedings of the
# AutoCarto 9 Conference, 462–71. Baltimore, MD, 1989.
# https://pdfs.semanticscholar.org/875e/12ce948012b3eced58c0f1470dba51ef87ef.pdf
#
# This script written by Paulo Raposo (pauloj.raposo [at] outlook.com) and Randall Brown
# (ranbrown8448 [at] gmail.com). Under MIT license (see LICENSE file).
#
# Dependencies:
#   - nvector (see https://pypi.python.org/pypi/nvector and http://www.navlab.net/nvector),
#   - OGR Python bindings (packaged with GDAL).

################
# Modified by Vgrid
################

import math
from shapely.geometry import Point, Polygon, LinearRing


def findCrossedMeridiansByLatitude(vert1, vert2, newLat):
    """For finding pair of meridians at which a great circle defined by two points crosses the given latitude."""

    # Credit to Chris Veness, https://github.com/chrisveness/geodesy.

    theta = math.radians(newLat)

    theta1 = math.radians(vert1[0])
    lamb1 = math.radians(vert1[1])
    theta2 = math.radians(vert2[0])
    lamb2 = math.radians(vert2[1])

    dlamb = lamb2 - lamb1

    x = math.sin(theta1) * math.cos(theta2) * math.cos(theta) * math.sin(dlamb)
    y = math.sin(theta1) * math.cos(theta2) * math.cos(theta) * math.cos(
        dlamb
    ) - math.cos(theta1) * math.sin(theta2) * math.cos(theta)
    z = math.cos(theta1) * math.cos(theta2) * math.sin(theta) * math.sin(dlamb)

    if z * z > x * x + y * y:
        print("Great circle doesn't reach latitude.")

    lambm = math.atan2(-y, x)
    dlambI = math.acos(z / math.sqrt(x * x + y * y))

    lambI1 = lamb1 + lambm - dlambI
    lambI2 = lamb1 + lambm + dlambI

    lon1 = (math.degrees(lambI1) + 540) % 360 - 180
    lon2 = (math.degrees(lambI2) + 540) % 360 - 180

    return lon1, lon2


def lonCheck(lon1, lon2, pointlon1, pointlon2):
    lesser, greater = sorted([pointlon1, pointlon2])
    if lon1 > lesser and lon1 < greater:
        return lon1
    else:
        return lon2


def GetMidpoint(vert1, vert2):
    midLat = (vert1[0] + vert2[0]) / 2
    midLon = (vert1[1] + vert2[1]) / 2
    return (float(midLat), float(midLon))


def constructGeometry(facet):
    """Accepting a list from this script that stores vertices, return a Shapely Polygon object."""
    if len(facet) == 5:
        # This is a triangle facet of format (vert,vert,vert,vert,orient)
        vertexTuples = facet[:4]
    if len(facet) == 6:
        # This is a rectangle facet of format (vert,vert,vert,vert,vert,northboolean)
        vertexTuples = facet[:5]

    # Create a LinearRing with the vertices
    ring = LinearRing(
        [(vT[1], vT[0]) for vT in vertexTuples]
    )  # sequence: lon, lat (x,y)

    # Create a Polygon from the LinearRing
    poly = Polygon(ring)
    return poly


def divideFacet(aFacet):
    """Will always return four facets, given one, rectangle or triangle."""

    # Important: For all facets, first vertex built is always the most south-then-west, going counter-clockwise thereafter.

    if len(aFacet) == 5:
        # This is a triangle facet.

        orient = aFacet[4]  # get the string expressing this triangle's orientation

        #       Cases, each needing subdivision:
        #                    ______           ___   ___
        #       |\      /|   \    /   /\     |  /   \  |     ^
        #       | \    / |    \  /   /  \    | /     \ |     N
        #       |__\  /__|     \/   /____\   |/       \|
        #
        #        up    up     down    up     down    down   -- orientations, as "u" or "d" in code below.

        # Find the geodetic bisectors of the three sides, store in sequence using edges defined
        # by aFacet vertex indeces: [0]&[1] , [1]&[2] , [2]&[3]
        newVerts = []

        for i in range(3):
            if aFacet[i][0] == aFacet[i + 1][0] or aFacet[i][1] == aFacet[i + 1][1]:
                newVerts.append(GetMidpoint(aFacet[i], aFacet[i + 1]))
            else:
                newLat = (aFacet[i][0] + aFacet[i + 1][0]) / 2
                newLon1, newLon2 = findCrossedMeridiansByLatitude(
                    aFacet[i], aFacet[i + 1], newLat
                )

                newLon = lonCheck(newLon1, newLon2, aFacet[i][1], aFacet[i + 1][1])

                newVert = (newLat, newLon)
                newVerts.append(newVert)

        if orient == "u":
            #          In the case of up facets, there will be one "top" facet
            #          and 3 "bottom" facets after subdivision; we build them in the sequence inside the triangles:
            #
            #                   2
            #                  /\         Outside the triangle, a number is the index of the vertex in aFacet,
            #                 / 1\        and a number with an asterisk is the index of the vertex in newVerts.
            #             2* /____\ 1*
            #               /\ 0  /\
            #              /2 \  /3 \
            #             /____\/____\
            #           0or3   0*     1

            newFacet0 = [newVerts[0], newVerts[1], newVerts[2], newVerts[0], "d"]
            newFacet1 = [newVerts[2], newVerts[1], aFacet[2], newVerts[2], "u"]
            newFacet2 = [aFacet[0], newVerts[0], newVerts[2], aFacet[0], "u"]
            newFacet3 = [newVerts[0], aFacet[1], newVerts[1], newVerts[0], "u"]

        if orient == "d":
            #          In the case of down facets, there will be three "top" facets
            #          and 1 "bottom" facet after subdivision; we build them in the sequence inside the triangles:
            #
            #            2_____1*_____1
            #             \ 2  /\ 3  /
            #              \  / 0\  /    Outside the triangle, a number is the index of the vertex in aFacet,
            #               \/____\/     and a number with an asterisk is the index of the vertex in newVerts.
            #              2*\ 1  /0*
            #                 \  /
            #                  \/
            #                 0or3

            newFacet0 = [newVerts[2], newVerts[0], newVerts[1], newVerts[2], "u"]
            newFacet1 = [aFacet[0], newVerts[0], newVerts[2], aFacet[0], "d"]
            newFacet2 = [newVerts[2], newVerts[1], aFacet[2], newVerts[2], "d"]
            newFacet3 = [newVerts[0], aFacet[1], newVerts[1], newVerts[0], "d"]

    if len(aFacet) == 6:
        # This is a rectangle facet.

        northBoolean = aFacet[5]  # true for north, false for south

        if northBoolean:
            # North pole rectangular facet.

            # Build new facets in the sequence inside the polygons:

            #          3..........2   <-- North Pole
            #           |        |
            #           |   1    |    Outside the polys, a number is the index of the vertex in aFacet,
            #           |        |    and a number with an asterisk is the index of the vertex in newVerts.
            #           |        |
            #         2*|--------|1*           /\
            #           |\      /|  on globe  /__\
            #           | \ 0  / |  -------> /\  /\
            #           |  \  /  |          /__\/__\
            #           | 2 \/ 3 |
            #       0or4''''''''''1
            #               0*

            newVerts = []

            for i in range(4):
                if i != 2:
                    # on iter == 1 we're going across the north pole - don't need this midpoint.

                    if (
                        aFacet[i][0] == aFacet[i + 1][0]
                        or aFacet[i][1] == aFacet[i + 1][1]
                    ):
                        newVerts.append(GetMidpoint(aFacet[i], aFacet[i + 1]))
                    else:
                        newLat = (aFacet[i][0] + aFacet[i + 1][0]) / 2
                        newLon1, newLon2 = findCrossedMeridiansByLatitude(
                            aFacet[i], aFacet[i + 1], newLat
                        )

                        newLon = lonCheck(
                            newLon1, newLon2, aFacet[i][1], aFacet[i + 1][1]
                        )

                        newVert = (newLat, newLon)
                        newVerts.append(newVert)

            newFacet0 = [
                newVerts[0],
                newVerts[1],
                newVerts[2],
                newVerts[0],
                "d",
            ]  # triangle
            newFacet1 = [
                newVerts[2],
                newVerts[1],
                aFacet[2],
                aFacet[3],
                newVerts[2],
                True,
            ]  # rectangle
            newFacet2 = [
                aFacet[0],
                newVerts[0],
                newVerts[2],
                aFacet[0],
                "u",
            ]  # triangle
            newFacet3 = [
                newVerts[0],
                aFacet[1],
                newVerts[1],
                newVerts[0],
                "u",
            ]  # triangle

        else:
            # South pole rectangular facet

            #               1*
            #          3..........2
            #           | 2 /\ 3 |     Outside the polys, a number is the index of the vertex in aFacet,
            #           |  /  \  |     and a number with an asterisk is the index of the vertex in newVerts.
            #           | / 0  \ |
            #           |/      \|           ________
            #         2*|--------|0*         \  /\  /
            #           |        |  on globe  \/__\/
            #           |   1    |  ------->   \  /
            #           |        |              \/
            #           |        |
            #       0or4'''''''''1   <-- South Pole

            newVerts = []

            for i in range(4):
                if i != 0:
                    # on iter == 3 we're going across the south pole - don't need this midpoint
                    if (
                        aFacet[i][0] == aFacet[i + 1][0]
                        or aFacet[i][1] == aFacet[i + 1][1]
                    ):
                        newVerts.append(GetMidpoint(aFacet[i], aFacet[i + 1]))
                    else:
                        newLat = (aFacet[i][0] + aFacet[i + 1][0]) / 2
                        newLon1, newLon2 = findCrossedMeridiansByLatitude(
                            aFacet[i], aFacet[i + 1], newLat
                        )

                        newLon = lonCheck(
                            newLon1, newLon2, aFacet[i][1], aFacet[i + 1][1]
                        )

                        newVert = newLat, newLon
                        newVerts.append(newVert)

            newFacet0 = [
                newVerts[2],
                newVerts[0],
                newVerts[1],
                newVerts[2],
                "u",
            ]  # triangle
            newFacet1 = [
                aFacet[0],
                aFacet[1],
                newVerts[0],
                newVerts[2],
                aFacet[0],
                False,
            ]  # rectangle
            newFacet2 = [
                newVerts[2],
                newVerts[1],
                aFacet[3],
                newVerts[2],
                "d",
            ]  # triangle
            newFacet3 = [
                newVerts[1],
                newVerts[0],
                aFacet[2],
                newVerts[1],
                "d",
            ]  # triangle

    # In all cases, return the four facets made in a list
    return [newFacet0, newFacet1, newFacet2, newFacet3]


################
# Added by Vgrid
################


def qtm_id_to_facet(qtm_id):
    # Base octahedral face definitions (these define the initial 8 triangular faces of QTM)
    p90_n180, p90_n90, p90_p0, p90_p90, p90_p180 = (
        (90.0, -180.0),
        (90.0, -90.0),
        (90.0, 0.0),
        (90.0, 90.0),
        (90.0, 180.0),
    )
    p0_n180, p0_n90, p0_p0, p0_p90, p0_p180 = (
        (0.0, -180.0),
        (0.0, -90.0),
        (0.0, 0.0),
        (0.0, 90.0),
        (0.0, 180.0),
    )
    n90_n180, n90_n90, n90_p0, n90_p90, n90_p180 = (
        (-90.0, -180.0),
        (-90.0, -90.0),
        (-90.0, 0.0),
        (-90.0, 90.0),
        (-90.0, 180.0),
    )

    initial_facets = {
        "1": [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
        "2": [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
        "3": [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
        "4": [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
        "5": [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
        "6": [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
        "7": [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
        "8": [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
    }

    base_facet = initial_facets.get(qtm_id[0])
    if base_facet is None:
        raise ValueError("Invalid QTM ID: Base facet must be 1-8")

    # Recursively follow the QTM ID to refine the facet
    facet = base_facet
    for level in range(1, len(qtm_id)):
        facet = divideFacet(facet)[int(qtm_id[level])]

    return facet


#   return facet[:-1]  # Drop the True/False flag


def latlon_to_qtm_id(lat, lon, resolution):
    # Base octahedral face definitions
    p90_n180, p90_n90, p90_p0, p90_p90, p90_p180 = (
        (90.0, -180.0),
        (90.0, -90.0),
        (90.0, 0.0),
        (90.0, 90.0),
        (90.0, 180.0),
    )
    p0_n180, p0_n90, p0_p0, p0_p90, p0_p180 = (
        (0.0, -180.0),
        (0.0, -90.0),
        (0.0, 0.0),
        (0.0, 90.0),
        (0.0, 180.0),
    )
    n90_n180, n90_n90, n90_p0, n90_p90, n90_p180 = (
        (-90.0, -180.0),
        (-90.0, -90.0),
        (-90.0, 0.0),
        (-90.0, 90.0),
        (-90.0, 180.0),
    )

    # Initial 8 facets
    initial_facets = {
        "1": [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
        "2": [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
        "3": [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
        "4": [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
        "5": [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
        "6": [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
        "7": [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
        "8": [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
    }

    # Find the initial facet containing (lat, lon)
    for facet_id, facet in initial_facets.items():
        facet_geom = constructGeometry(facet)
        if facet_geom.contains(Point(lon, lat)):
            qtm_id = facet_id
            current_facet = facet
            break
    else:
        raise ValueError("Point is outside the valid range")

    # Refine the facet through subdivisions up to the given resolution
    for _ in range(1, resolution):
        subfacets = divideFacet(current_facet)
        for i, subfacet in enumerate(subfacets):
            subfacet_geom = constructGeometry(subfacet)
            if subfacet_geom.contains(Point(lon, lat)):
                qtm_id += str(i)
                current_facet = subfacet
                break

    return qtm_id


def qtm_id_to_latlon(qtm_id):
    # Retrieve the facet corresponding to this QTM ID
    facet = qtm_id_to_facet(qtm_id)

    # Exclude the last element if it's a string (e.g., 'a', 'b', 'c', 'd')
    coords = facet[:-1] if isinstance(facet[-1], str) else facet

    # Calculate the average latitude and longitude
    latitudes = [lat for lat, lon in coords]
    longitudes = [lon for lat, lon in coords]
    avg_latitude = sum(latitudes) / len(latitudes)
    avg_longitude = sum(longitudes) / len(longitudes)

    return avg_latitude, avg_longitude


def qtm_parent(qtm_id):
    if len(qtm_id) <= 1:
        return None  # Root facets (1-8) have no parent
    return qtm_id[:-1]


def qtm_children(qtm_id, resolution=None):
    if resolution is None:
        resolution = len(qtm_id) + 1  # Default to next level

    children = []

    def recurse(current_id, current_resolution):
        if current_resolution == resolution:
            children.append(current_id)
            return
        for i in range(4):  # Subdivide into 4 sub-triangles
            recurse(current_id + str(i), current_resolution + 1)

    recurse(qtm_id, len(qtm_id))
    return children



import argparse
import json
from shapely.wkt import loads
import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat
    from vgrid.generator.settings import isea4t_res_accuracy_dict

from shapely.ops import unary_union
from tqdm import tqdm
from shapely.geometry import Polygon, box, shape
from vgrid.utils.antimeridian import fix_polygon
from vgrid.generator.settings import (
    max_cells,
    isea4t_base_cells,
    geodesic_dggs_to_feature,
)


def fix_isea4t_wkt(isea4t_wkt):
    # Extract the coordinate section
    coords_section = isea4t_wkt[isea4t_wkt.index("((") + 2 : isea4t_wkt.index("))")]
    coords = coords_section.split(",")
    # Append the first point to the end if not already closed
    if coords[0] != coords[-1]:
        coords.append(coords[0])
    fixed_coords = ", ".join(coords)
    return f"POLYGON (({fixed_coords}))"


def fix_isea4t_antimeridian_cells(isea4t_boundary, threshold=-100):
    """
    Adjusts polygon coordinates to handle antimeridian crossings.
    """
    lon_lat = [(float(lon), float(lat)) for lon, lat in isea4t_boundary.exterior.coords]

    if any(lon < threshold for lon, _ in lon_lat):
        adjusted_coords = [(lon - 360 if lon > 0 else lon, lat) for lon, lat in lon_lat]
    else:
        adjusted_coords = lon_lat

    return Polygon(adjusted_coords)


def isea4t_cell_to_polygon(isea4t_dggs, isea4t_cell):
    cell_to_shp = isea4t_dggs.convert_dggs_cell_outline_to_shape_string(
        isea4t_cell, ShapeStringFormat.WKT
    )
    cell_to_shp_fixed = fix_isea4t_wkt(cell_to_shp)
    cell_polygon = loads(cell_to_shp_fixed)
    return cell_polygon


def get_isea4t_children_cells(isea4t_dggs, base_cells, target_resolution):
    """
    Recursively generate DGGS cells for the desired resolution.
    """
    current_cells = base_cells
    for res in range(target_resolution):
        next_cells = []
        for cell in current_cells:
            children = isea4t_dggs.get_dggs_cell_children(DggsCell(cell))
            next_cells.extend([child._cell_id for child in children])
        current_cells = next_cells
    return current_cells


def get_isea4t_children_cells_within_bbox(
    isea4t_dggs, bounding_cell, bbox, target_resolution
):
    current_cells = [
        bounding_cell
    ]  # Start with a list containing the single bounding cell
    bounding_resolution = len(bounding_cell) - 2

    for res in range(bounding_resolution, target_resolution):
        next_cells = []
        for cell in current_cells:
            # Get the child cells for the current cell
            children = isea4t_dggs.get_dggs_cell_children(DggsCell(cell))
            for child in children:
                # Convert child cell to geometry
                child_shape = isea4t_cell_to_polygon(isea4t_dggs, child)
                if child_shape.intersects(bbox):
                    # Add the child cell ID to the next_cells list
                    next_cells.append(child._cell_id)
        if not next_cells:  # Break early if no cells remain
            break
        current_cells = (
            next_cells  # Update current_cells to process the next level of children
        )

    return current_cells


def generate_grid(isea4t_dggs, resolution):
    # accuracy = isea4t_res_accuracy_dict.get(resolution)
    children = get_isea4t_children_cells(isea4t_dggs, isea4t_base_cells, resolution)
    isea4t_features = []
    for child in tqdm(children, desc="Generating ISEA4T DGGS", unit=" cells"):
        isea4t_cell = DggsCell(child)
        cell_polygon = isea4t_cell_to_polygon(isea4t_dggs, isea4t_cell)
        isea4t_id = isea4t_cell.get_cell_id()
        num_edges = 3
        if resolution == 0:
            cell_polygon = fix_polygon(cell_polygon)
        elif (
            isea4t_id.startswith("00")
            or isea4t_id.startswith("09")
            or isea4t_id.startswith("14")
            or isea4t_id.startswith("04")
            or isea4t_id.startswith("19")
        ):
            cell_polygon = fix_isea4t_antimeridian_cells(cell_polygon)

        isea4t_feature = geodesic_dggs_to_feature(
            "isea4t", isea4t_id, resolution, cell_polygon, num_edges
        )
        isea4t_features.append(isea4t_feature)

    return {"type": "FeatureCollection", "features": isea4t_features}


def generate_grid_within_bbox(isea4t_dggs, resolution, bbox):
    accuracy = isea4t_res_accuracy_dict.get(resolution)
    bounding_box = box(*bbox)
    bounding_box_wkt = bounding_box.wkt  # Create a bounding box polygon
    isea4t_shapes = isea4t_dggs.convert_shape_string_to_dggs_shapes(
        bounding_box_wkt, ShapeStringFormat.WKT, accuracy
    )
    isea4t_shape = isea4t_shapes[0]
    bbox_cells = isea4t_shape.get_shape().get_outer_ring().get_cells()
    bounding_cell = isea4t_dggs.get_bounding_dggs_cell(bbox_cells)
    bounding_children = get_isea4t_children_cells_within_bbox(
        isea4t_dggs, bounding_cell.get_cell_id(), bounding_box, resolution
    )
    isea4t_features = []
    for child in tqdm(bounding_children, desc="Generating ISEA4T DGGS", unit=" cells"):
        isea4t_cell = DggsCell(child)
        cell_polygon = isea4t_cell_to_polygon(isea4t_dggs, isea4t_cell)
        isea4t_id = isea4t_cell.get_cell_id()
        if resolution == 0:
            cell_polygon = fix_polygon(cell_polygon)

        elif (
            isea4t_id.startswith("00")
            or isea4t_id.startswith("09")
            or isea4t_id.startswith("14")
            or isea4t_id.startswith("04")
            or isea4t_id.startswith("19")
        ):
            cell_polygon = fix_isea4t_antimeridian_cells(cell_polygon)
        num_edges = 3

        # if cell_polygon.intersects(bounding_box):
        isea4t_feature = geodesic_dggs_to_feature(
            "isea4t", isea4t_id, resolution, cell_polygon, num_edges
        )
        isea4t_features.append(isea4t_feature)

    return {"type": "FeatureCollection", "features": isea4t_features}


def generate_grid_resample(isea4t_dggs, resolution, geojson_features):
    accuracy = isea4t_res_accuracy_dict.get(resolution)
    # Step 1: Unify all geometries into a single shape
    geometries = [
        shape(feature["geometry"]) for feature in geojson_features["features"]
    ]
    unified_geom = unary_union(geometries)
    unified_geom_wkt = unified_geom.wkt

    # Step 2: Generate DGGS shapes from WKT geometry
    isea4t_shapes = isea4t_dggs.convert_shape_string_to_dggs_shapes(
        unified_geom_wkt, ShapeStringFormat.WKT, accuracy
    )
    isea4t_shape = isea4t_shapes[0]
    bbox_cells = isea4t_shape.get_shape().get_outer_ring().get_cells()
    bounding_cell = isea4t_dggs.get_bounding_dggs_cell(bbox_cells)

    # Step 3: Generate children cells within geometry bounds
    bounding_children = get_isea4t_children_cells_within_bbox(
        isea4t_dggs, bounding_cell.get_cell_id(), unified_geom, resolution
    )

    isea4t_features = []
    for child in tqdm(bounding_children, desc="Generating ISEA4T DGGS", unit=" cells"):
        isea4t_cell = DggsCell(child)
        cell_polygon = isea4t_cell_to_polygon(isea4t_dggs, isea4t_cell)
        isea4t_id = isea4t_cell.get_cell_id()

        if resolution == 0:
            cell_polygon = fix_polygon(cell_polygon)
        elif isea4t_id.startswith(("00", "09", "14", "04", "19")):
            cell_polygon = fix_isea4t_antimeridian_cells(cell_polygon)

        num_edges = 3

        # Optional: only include cells intersecting original geometry
        if not cell_polygon.intersects(unified_geom):
            continue

        isea4t_feature = geodesic_dggs_to_feature(
            "isea4t", isea4t_id, resolution, cell_polygon, num_edges
        )
        isea4t_features.append(isea4t_feature)

    return {"type": "FeatureCollection", "features": isea4t_features}


def main():
    parser = argparse.ArgumentParser(description="Generate Open-Eaggr ISEA4T DGGS.")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="Resolution [0..25]"
    )
    # Resolution max range: [0..39]
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )

    if platform.system() == "Windows":
        isea4t_dggs = Eaggr(Model.ISEA4T)
        args = parser.parse_args()
        resolution = args.resolution
        bbox = args.bbox if args.bbox else [-180, -90, 180, 90]
        if resolution < 0 or resolution > 25:
            print("Please select a resolution in [0..25] range and try again ")
            return

        if bbox == [-180, -90, 180, 90]:
            total_cells = 20 * (4**resolution)
            print(f"Resolution {resolution} will generate {total_cells} cells ")
            if total_cells > max_cells:
                print(f"which exceeds the limit of {max_cells}.")
                print("Please select a smaller resolution and try again.")
                return

            geojson_features = generate_grid(isea4t_dggs, resolution)

        else:
            geojson_features = generate_grid_within_bbox(isea4t_dggs, resolution, bbox)

        # Define the GeoJSON file path
        geojson_path = f"isea4t_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



from shapely.geometry import shape, Polygon
import argparse
import json
from vgrid.utils import qtm
from vgrid.generator.settings import geodesic_dggs_to_feature
from shapely.ops import unary_union
from tqdm import tqdm

p90_n180, p90_n90, p90_p0, p90_p90, p90_p180 = (
    (90.0, -180.0),
    (90.0, -90.0),
    (90.0, 0.0),
    (90.0, 90.0),
    (90.0, 180.0),
)
p0_n180, p0_n90, p0_p0, p0_p90, p0_p180 = (
    (0.0, -180.0),
    (0.0, -90.0),
    (0.0, 0.0),
    (0.0, 90.0),
    (0.0, 180.0),
)
n90_n180, n90_n90, n90_p0, n90_p90, n90_p180 = (
    (-90.0, -180.0),
    (-90.0, -90.0),
    (-90.0, 0.0),
    (-90.0, 90.0),
    (-90.0, 180.0),
)


def generate_grid(resolution):
    levelFacets = {}
    QTMID = {}

    for lvl in tqdm(range(resolution), desc="Generating QTM DGGS"):
        levelFacets[lvl] = []
        QTMID[lvl] = []
        qtm_features = []  # Store GeoJSON features separately

        if lvl == 0:
            initial_facets = [
                [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
                [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
                [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
                [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
                [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
                [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
                [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
                [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
            ]

            for i, facet in enumerate(initial_facets):
                facet_geom = qtm.constructGeometry(facet)
                QTMID[0].append(str(i + 1))
                levelFacets[0].append(facet)
                qtm_id = QTMID[0][i]
                num_edges = 3
                qtm_feature = geodesic_dggs_to_feature(
                    "qtm", qtm_id, resolution, facet_geom, num_edges
                )
                qtm_features.append(qtm_feature)

        else:
            for i, pf in enumerate(levelFacets[lvl - 1]):
                subdivided_facets = qtm.divideFacet(pf)
                for j, subfacet in enumerate(subdivided_facets):
                    new_id = QTMID[lvl - 1][i] + str(j)
                    QTMID[lvl].append(new_id)
                    levelFacets[lvl].append(subfacet)
                    if lvl == resolution - 1:
                        subfacet_geom = qtm.constructGeometry(subfacet)
                        qtm_id = new_id
                        num_edges = 3
                        qtm_feature = geodesic_dggs_to_feature(
                            "qtm", qtm_id, resolution, subfacet_geom, num_edges
                        )
                        qtm_features.append(qtm_feature)
    return {"type": "FeatureCollection", "features": qtm_features}


def generate_grid_within_bbox(resolution, bbox):
    """Generates a Dutton QTM grid at a specific resolution within a bounding box and saves it as GeoJSON."""
    levelFacets = {}
    QTMID = {}
    qtm_features = []

    # Convert bbox to Polygon
    bbox_poly = Polygon(
        [
            (bbox[0], bbox[1]),  # min_lon, min_lat
            (bbox[2], bbox[1]),  # max_lon, min_lat
            (bbox[2], bbox[3]),  # max_lon, max_lat
            (bbox[0], bbox[3]),  # min_lon, max_lat
            (bbox[0], bbox[1]),  # Close the polygon
        ]
    )

    for lvl in tqdm(range(resolution), desc="Generating QTM DGGS"):
        levelFacets[lvl] = []
        QTMID[lvl] = []

        if lvl == 0:
            initial_facets = [
                [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
                [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
                [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
                [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
                [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
                [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
                [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
                [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
            ]

            for i, facet in enumerate(initial_facets):
                QTMID[0].append(str(i + 1))
                facet_geom = qtm.constructGeometry(facet)
                levelFacets[0].append(facet)
                if shape(facet_geom).intersects(bbox_poly) and resolution == 1:
                    qtm_id = QTMID[0][i]
                    num_edges = 3
                    qtm_feature = geodesic_dggs_to_feature(
                        "qtm", qtm_id, resolution, facet_geom, num_edges
                    )
                    qtm_features.append(qtm_feature)
        else:
            for i, pf in enumerate(levelFacets[lvl - 1]):
                subdivided_facets = qtm.divideFacet(pf)
                for j, subfacet in enumerate(subdivided_facets):
                    subfacet_geom = qtm.constructGeometry(subfacet)
                    if shape(subfacet_geom).intersects(
                        bbox_poly
                    ):  # Only keep intersecting facets
                        new_id = QTMID[lvl - 1][i] + str(j)
                        QTMID[lvl].append(new_id)
                        levelFacets[lvl].append(subfacet)
                        if (
                            lvl == resolution - 1
                        ):  # Only store final resolution in GeoJSON
                            qtm_id = new_id
                            num_edges = 3
                            qtm_feature = geodesic_dggs_to_feature(
                                "qtm", qtm_id, resolution, subfacet_geom, num_edges
                            )
                            qtm_features.append(qtm_feature)
    return {"type": "FeatureCollection", "features": qtm_features}


def generate_grid_resample(resolution, geojson_features):
    """Generates a Dutton QTM grid at a specific resolution within geojson_features and returns it as GeoJSON."""
    levelFacets = {}
    QTMID = {}
    qtm_features = []

    # Step 1: Union all input GeoJSON geometries
    geometries = [
        shape(feature["geometry"]) for feature in geojson_features["features"]
    ]
    unified_geom = unary_union(geometries)

    for lvl in tqdm(range(resolution), desc="Generating QTM DGGS"):
        levelFacets[lvl] = []
        QTMID[lvl] = []

        if lvl == 0:
            initial_facets = [
                [p0_n180, p0_n90, p90_n90, p90_n180, p0_n180, True],
                [p0_n90, p0_p0, p90_p0, p90_n90, p0_n90, True],
                [p0_p0, p0_p90, p90_p90, p90_p0, p0_p0, True],
                [p0_p90, p0_p180, p90_p180, p90_p90, p0_p90, True],
                [n90_n180, n90_n90, p0_n90, p0_n180, n90_n180, False],
                [n90_n90, n90_p0, p0_p0, p0_n90, n90_n90, False],
                [n90_p0, n90_p90, p0_p90, p0_p0, n90_p0, False],
                [n90_p90, n90_p180, p0_p180, p0_p90, n90_p90, False],
            ]

            for i, facet in enumerate(initial_facets):
                QTMID[0].append(str(i + 1))
                facet_geom = qtm.constructGeometry(facet)
                levelFacets[0].append(facet)

                if shape(facet_geom).intersects(unified_geom) and resolution == 1:
                    qtm_id = QTMID[0][i]
                    num_edges = 3
                    qtm_feature = geodesic_dggs_to_feature(
                        "qtm", qtm_id, resolution, facet_geom, num_edges
                    )
                    qtm_features.append(qtm_feature)

        else:
            for i, pf in enumerate(levelFacets[lvl - 1]):
                subdivided_facets = qtm.divideFacet(pf)
                for j, subfacet in enumerate(subdivided_facets):
                    subfacet_geom = qtm.constructGeometry(subfacet)
                    if shape(subfacet_geom).intersects(unified_geom):
                        new_id = QTMID[lvl - 1][i] + str(j)
                        QTMID[lvl].append(new_id)
                        levelFacets[lvl].append(subfacet)

                        if lvl == resolution - 1:
                            qtm_id = new_id
                            num_edges = 3
                            qtm_feature = geodesic_dggs_to_feature(
                                "qtm", qtm_id, resolution, subfacet_geom, num_edges
                            )
                            qtm_features.append(qtm_feature)

    return {"type": "FeatureCollection", "features": qtm_features}


def main():
    parser = argparse.ArgumentParser(description="Generate QTM DGGS.")
    parser.add_argument(
        "-r", "--resolution", required=True, type=int, help="Resolution [1..24]."
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )

    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox if args.bbox else [-180, -90, 180, 90]

    if resolution < 1 or resolution > 24:
        print("Please select a resolution in [1..24] range and try again ")
        return

    if bbox == [-180, -90, 180, 90]:
        geojson_features = generate_grid(resolution)

    else:
        geojson_features = generate_grid_within_bbox(resolution, bbox)

    if geojson_features:
        geojson_path = f"qtm_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""


## Defines a DGGS cell.
class DggsCell(object):
    ## Sets the cell ID of the DGGS cell.
    #  @param cell_id ID of the cell as a string.
    def __init__(self, cell_id):
        self._cell_id = cell_id

    ## @return ID of the cell as a string.
    def get_cell_id(self):
        return self._cell_id

    ## @return True if cells have the same ID, otherwise false.
    def __eq__(self, other):
        return self._cell_id == other.get_cell_id()

    ## @return Opposite of the == operator.
    def __ne__(self, other):
        return not (self == other)
